/***********************************************************************
|  Genealogy management system
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* command.c: interpret a command */

#include "structs.h"

static long sibsetup()
{
long prs2;
nl();
prs2 = getdad(cur_prs_rowid);
if(!prs2)
prs2 = getmom(cur_prs_rowid);
if(!prs2) {
printf("%s has no father or mother\n", cur_fname);
return 0L;
}
return prs2;
} /* sibsetup */

command()
{
char cmd, *relation;
int n, i, adoptflag = 0;
long prs2;
static char *relnames[] = {"father", "mother", "children", "husband", "wife",
"sibling", "younger siblings", "older siblings"};
static char relcodes[] = "FMCHWSYO";
char digits[10];
long atol();

top:
cmd = promptfor("Action:", "+>pmfhwcsyoPMFHWCS123456789?tIuZ#^");
top2:
n = 0;
prs2 = cur_prs_rowid;
relation = strchr(relcodes, toupper(cmd));
if(relation)
relation = relnames[relation - relcodes];
if(isdigit(cmd)) relation = relnames[2];

switch(cmd) {
case '#':
printf("\nserial number = %ld\n", cur_prs_rowid);
gets(line);
if(!line[0]) goto top;
set_curperson(atol(line));
break;

case '^':
printf("\n\nGenerate formatted output file, appropriate for review or printing\n");
printgo(1);
nl();
goto top;

case 's':
printf("\nUse either `o' or `y', for an older or younger sibling.\n");
goto top;

case '>':
adoptflag = 1;
cmd = promptfor("", "cmfCMF123456789");
goto top2;

case 'Z': zap_person(); break;

case 'I': nl(); instruct(); break;

case 'p': nl(); set_curperson(searchfor(NULL)); break;

case 't': nl(); text_interactive(); break;

case 'u':
cmd = promptfor("pdate:  names, dates, shortnotes, gender?", "ndsg");
nl();
nl();
if(cmd == 's') up_shortnote();
if(cmd == 'd') up_date();
if(cmd == 'n') up_name();
if(cmd == 'g') up_gender();
break;

case '?':
printf("\nhwfmc   Move to husband, wife, father, mother, child.\n");
printf("oy      Move to an older or younger sibling.\n");
printf("1-9     Move to one of the children, ordered by birth.\n");
printf("p       Move to a new (possibly unrelated) person.\n");
printf("HWFMCS  Add a husband, wife, father, mother, child, or sibling.\n");
printf("P       Add an unrelated person.\n");
printf(">       Activate adoption/caregiving mode.\n");
printf("q       quit this program.\n");
printf(".       abort any data entry in progress and return to the main screen.\n");
printf("?       Print a help synopsis, the one you are currently reading.\n");
printf("I       Read the instructions for this program.\n");
printf("t       Review, update, or add text entries for %s.\n", cur_fname);
printf("u       Update shortnotes, names, or dates for %s.\n", cur_fname);
printf("+       Merge %s with another person; the two become one.\n", cur_fname);
printf("Z       Zap (remove) %s's links from the database.\n", cur_fname);
nl();
goto top;

case '1': case '2': case '3': case '4': case '5':
case '6': case '7': case '8': case '9':
n = cmd-'0';
cmd = 'c';
case 'c': case 'h': case 'w':
adoptflag = 0;
 case 'm': case 'f':
spreadout(cur_prs_rowid, cmd);
if(!adoptflag) {
if(!spreadnum || spreaddata[0].adopt && cmd != 'c') {
printf("\n%s has no %s\n", cur_fname, relation);
goto top;
}
if(n > spreadnum) {
printf("\n%s does not have %d children\n", cur_fname, n);
goto top;
}
if(!n) {
if(spreadnum == 1 || cmd == 'm' || cmd == 'f') n = 1;
else {
printf("#");
strcpy(digits, "123456789");
if(spreadnum < 9) digits[spreadnum] = 0;
n = getletter(digits) - '0';
}
}
} else {
/* type is m or f, and adoptflag is set */
n = spreadnum;
if(n && !spreaddata[0].adopt) --n;
if(!n) {
printf("\n%s has no caregiving %s\n", cur_fname, relation);
adoptflag = 0;
goto top;
}
if(n > 1) {
printf("#");
strcpy(digits, "123456789");
if(n < 9) digits[n] = 0;
n = getletter(digits) - '0';
}
if(!spreaddata[0].adopt) ++n;
adoptflag = 0;
}
set_curperson(spreaddata[n-1].prsid);
nl();
break;

case 'y': case 'o':
if(!(prs2 = sibsetup())) goto top;
spreadout(prs2, 'c');
for(i=0; i<spreadnum; ++i)
if(spreaddata[i].prsid == cur_prs_rowid) break;
if(i == spreadnum)
eprint("child/parent/child link not found");
if(cmd == 'o') {
while(--i >= 0)
if(!spreaddata[i].adopt) break;
} else {
while(++i < spreadnum)
if(!spreaddata[i].adopt) break;
}
if(i < 0 || i >= spreadnum) {
printf("%s has no %s\n", cur_fname, relation);
goto top;
}
set_curperson(spreaddata[i].prsid);
break;

/* add person */
case 'S':
/* jump up to father or mother, and run 'C' */
if(!(prs2 = sibsetup())) goto top;

case 'H': case 'W': case 'M': case 'F': case 'C':
if(cmd == 'S') cmd = 'C';
else nl();
cmd = tolower(cmd);
if(cur_prs.sex == 'm' && cmd == 'h') {
printf("%s is male and cannot have a husband\n", cur_fname);
goto top;
}
if(cur_prs.sex == 'f' && cmd == 'w') {
printf("%s is female and cannot have a wife\n", cur_fname);
goto top;
}
spreadout(prs2, cmd);
if(spreadnum &&
(cmd == 'm' || cmd == 'f') &&
!(spreaddata[0].adopt | adoptflag)) {
printf("%s already has a %s\n", cur_fname, relation);
goto top;
}
if(spreadnum == SPREADMAX)
eprint("too many links, limit %d\n", SPREADMAX);

case 'P':
n = addperson(prs2, cmd, cur_prs_rowid, relation, adoptflag);
adoptflag = 0;
if(n) goto top;
break;

case '+':
merge_current();
break;

default:
eprint("not yet implemented");
} /* end switch */
} /* command */

addperson(prs1, linktype, exception, relation, adoptflag)
long prs1, exception;
char linktype;
char *relation;
{
struct name *n;
struct qknote bp; /* specifically ask for birthplace */
struct person new_p, p1;
struct prslink l;
char c, deflast;
long prs2, new_prs2;
char fname[NAMELEN], lname[NAMELEN];
char *prs2sex;
extern long i_count;

new_prs2 = prs2 = 0;
deflast = 0;
lname[0] = 0;
linktype = tolower(linktype);

if(linktype != 'p') {
/* get data on prs1 */
birthname(prs1);
strcpy(fname, fld1);
strcpy(lname, fld3);
fname[20] = 0;
select_copy_person(prs1, &p1);
}
linktype_dates(&p1, fname, linktype, adoptflag);
nl();

if(linktype == 'f' || linktype == 'c' && p1.sex == 'm')
deflast = 1;
if(adoptflag) deflast = 0;
n = getname(deflast ? "first and middle name" : "full name", 1, deflast);
n->birth = 1;
if(n->search) {
if(linktype == 'p') {
printf("The search capability cannot be invoked when adding a new unrelated person\n\n");
return 1;
}

prs2 = searchfor(n);
if(prs2 == exception) {
printf("cannot link %s to %s, same person\n\n", fname, fname);
return 1;
}

birthname(prs2); /* for error messages */

/* if we're not careful, linking to an existing child might
 * create someone with two mothers. */
if(linktype == 'c' && !adoptflag) {
exception = p1.sex == 'm' ? getdad(prs2) : getmom(prs2);
if(exception) {
relation = p1.sex == 'm' ? "father" : "mother";
printf("%s already has a %s, cannot link %s as a second %s\n",
fld1, relation, fname, relation);
return 1;
}
}

select_copy_person(prs2, &new_p);

/* retrieved from the database, need to check sex and birthdate */
prs2sex = new_p.sex == 'm' ? "male" : "female";
if((new_p.sex == p1.sex && (linktype == 'h' || linktype == 'w')) ||
(linktype == new_p.sex && (linktype == 'f' || linktype == 'm'))) {
printf("%s is %s and cannot be %s's %s%s\n",
fld1, prs2sex, fname, adoptflag ? "caregiving " : "", relation);
return 1;
}

if(lowdate && new_p.birth &&
(new_p.birth < lowdate || new_p.birth > highdate)) {
printf("%s's birth%s\n", fld1, datemsg);
return 1;
}

} else { /* search for vs. create */

if(!deflast && !strcmp(lname, fld3)) {
printf("\nCoincidence discovered, both %s and %s have the same last name %s.\n",
fname, fld1, lname);
printf("Each person should be entered using his or her birth or maiden name.\n");
/* whose name should be changed?
 * If adding a mother or wife, the new person's name is in question.
 * If adopting a child, the new person's name is in question.
 * If anything else (adding husband or adopting parent),
 * my name is in question. */
if(linktype == 'm' && !adoptflag ||
linktype == 'w' ||
linktype == 'c' && adoptflag)
deflast = 1;
else
deflast = 0;
printf("is %s's birth name really %s?",
deflast ? fld1 : fname, lname);
if(!get_yn("")) {
do {
getline("original last name", "name", "L'-", "letters and spaces, or ' -", 0);
} while(!line[0]);
smartmix(line, 0);
nl();
if(deflast) {
change_lname(n, line);
} else {
/* the last one was easy, this time we have to update the database. */
struct name old_n;
select_copy_name(birth_rowid, &old_n);
change_lname(&old_n, line);
insupd_name(birth_rowid, &old_n);
set_curperson(cur_prs_rowid);
}
}
}
splitstring(n->fullname);

init_0(&new_p);
new_p.birth = getdate("birthdate", datemsg, 0);
if(new_p.birth) {
splitstring(n->fullname);
lifetime_dates(&new_p, fld1, 0, 120);
new_p.death = getdate("deathdate", datemsg, 0);
}

switch(linktype) { /* establish sex */
case 'f': case 'h': c = 'm'; break;
case 'm': case 'w': c = 'f'; break;
default: c = promptfor("Gender: ", "mMfF"); nl(); break;
} /* end switch on link type */
new_p.sex = tolower(c);

getqknote('b');
init_0(&bp);
strcpy(bp.note, line);
bp.ntype = 'b';
} /* found/created prs2 */

init_0(&l);
if(linktype == 'h' || linktype == 'w') {
lifetime_dates(&p1, fname, 10, 120);
if(new_p.birth > lowdate) lowdate = new_p.birth;
if(new_p.death && new_p.death < highdate) highdate = new_p.death;
l.start = getdate("wedding date", "date is not within the couple's common lifetime", 0);
if(l.start) {
lowdate = l.start;
l.end = getdate("date of divorce", "date does not lie within the span of the couple's marriage", 0);
}
}

if(adoptflag) {
if(linktype == 'c') {
splitstring(n->fullname);
lifetime_dates(&new_p, fld1, 0, 120);
} else {
lifetime_dates(&p1, fname, 0, 120);
}
l.end = getdate("date caregiving relationship begins", datemsg, 0);
} /* date of adoption */

if(!prs2) {
new_p.txtc1 = 1;
new_p.txtcnt = 0;
prs2 = insupd_person(NULLKEY, &new_p);
new_prs2 = prs2;
n->prsid = prs2;
insupd_name(NULLKEY, n);
/*  This doesn't really work:
if(i_count >= 25)
printf("\nExcessive index clustering, you may wish to run the command `gng reindex'\n\n");
*/
if(bp.note[0]) {
bp.prsid = prs2;
insupd_qknote(NULLKEY, &bp);
}
}

set_curperson(prs2);

if(linktype != 'p') {
l.linktype = linktype; /* a good guess, for a start */
switch(linktype) {
case 'c':
l.linktype = p1.sex;
l.linktype = 'f'+'m' - l.linktype;
case 'w':
if(linktype == 'w') l.linktype = 'h';
l.prs1 = prs1;
l.prs2 = prs2;
break;
case 'm': case 'f':
case 'h':
l.prs1 = prs2;
l.prs2 = prs1;
break;
} /* end switch on link type */
l.adopt = adoptflag;

insupd_prslink(NULLKEY, &l);
nl();

/* build any implied links */
implied_links(&l);
}

#ifndef MSDOS
if(new_prs2 == 1000) mailstring("large", "Database is large");
if(new_prs2 == 10000) mailstring("huge", "Database is huge");
#endif
return 0; /* ok */
} /* addperson */

/* indentation string, shift text over */
static char indstring[21];

/* display short notes, to screen or file */
sn_display(of, rownums)
FILE *of; /* output file */
long *rownums;
{
int i, cnt;
char *sn, ntype;

cnt = 0;
for(i=0; i<QKNOTE_CAT; ++i) {
ntype = qkcat[i];
if(rownums) rownums[i] = 0;
sn = shortnote(cur_prs_rowid, ntype);
if(sn) {
if(rownums) rownums[i] = birth_rowid;
fprintf(of, "%s%s: %s\n",
indstring, qkdesc(ntype, 0), sn);
++cnt;
}
}

return cnt;
} /* sn_display */

link_summary(of, title, linktype, mustadopt)
FILE *of; /* output file */
char *title;
char linktype;
{
int i, l, cnt;
char s[80+25+NAMELEN];

strcpy(s, indstring);
strcat(s, title);
spreadout(cur_prs_rowid, linktype);
cnt = 0;

for(i=0; i<spreadnum; ++i) {
if(mustadopt && !spreaddata[i].adopt) continue;
if(cnt) strcat(s, ",");
++cnt;
l = strlen(s);
birthname(spreaddata[i].prsid);
strcat(s, " ");
if(spreaddata[i].adopt)
strcat(s, ">");
strcat(s, fld1);
if(linktype != 'c' && ci_strncmp(fld3, "maiden", 6)) {
strcat(s, " ");
strcat(s, fld3);
}
if((linktype == 'h' || linktype == 'w') &&
spreaddata[i].order) {
sprintf(s+strlen(s), " %04d", (int)(spreaddata[i].order / YEAR_DAY_RATIO));
}
if(strlen(s) > 78) {
s[l] = 0;
fprintf(of, "%s\n", s);
strcpy(s, indstring);
strcat(s, "  ");
strcat(s, s+l+1);
}
}

if(cnt) fprintf(of, "%s\n", s);
} /* link_summary */

topscreen(of, indlevel)
FILE *of; /* output file */
{
char s[80+NAMELEN];
char lname[NAMELEN];
long parent;
int i;
static char indchars[] = "-=+:#";

if(of == stdout) nl();

for(i=0; i<indlevel%10; ++i) {
indstring[2*i] = indchars[indlevel/10];
indstring[2*i+1] = ' ';
}
indstring[2*i] = 0;
fprintf(of, "%s%s %s %s",
indstring, cur_fname, cur_mname, cur_lname);
for(i=0; indstring[i]; ++i)
indstring[i] = ' ';

if(cur_prs.birth) {
fprintf(of, " %s", datestring(cur_prs.birth));
if(cur_prs.death) fprintf(of, "-%s", datestring(cur_prs.death));
}
fprintf(of, "\n");

if(of == stdout) {
if(cur_birthplace[0]) printf("Born: %s\n", cur_birthplace);
}

/* find parents */
s[0] = 0;
lname[0] = 0;
if(parent = getdad(cur_prs_rowid)) {
birthname(parent);
strcpy(s, fld1);
strcpy(lname, fld3);
} else strcpy(s, "???");

if(parent = getmom(cur_prs_rowid)) {
birthname(parent);
if(!lname[0])
strcpy(lname, fld3);
strcat(s, " and ");
strcat(s, fld1);
} else strcat(s, " and ???");

if(lname[0]) {
s[80] = 0;
strcat(s, " ");
strcat(s, lname);
fprintf(of, "%sParents: %s\n", indstring, s);
}

link_summary(of, "Caregiving fathers:", 'f', 1);
link_summary(of, "Caregiving mothers:", 'm', 1);
link_summary(of, "Married to:", (cur_prs.sex == 'm') ? 'w' : 'h', 0);
link_summary(of, "Children:", 'c', 0);

if(of != stdout)
sn_display(of, NULL);

if(!cur_prs.txtc1) { /* count text entries */
long textcount;
date firstyear;
textstats(cur_prs_rowid, &textcount, &firstyear);
cur_prs.txtcnt = textcount;
cur_prs.txtc1 = 1;
insupd_person(cur_prs_rowid, &cur_prs);
}
if(cur_prs.txtcnt)
fprintf(of, "%s%u text entries present.\n", indstring, cur_prs.txtcnt);

fprintf(of, "\n");
} /* topscreen */

/* variables for generating printable files.
 * These files contain the summary screens (generated by topscreen())
 * For a person's ancestors, a person's descendents, or a chain of people. */

static FILE *pf;
static int indlevel;
static long chainlink[201];
static int chainlength;
static long chaintop;

/* ancestral print */
ancprint(prsid)
long prsid;
{
long parent;

if(indlevel > 50) return; /* no more than 50 generations */
set_curperson(prsid);
topscreen(pf, indlevel);

++indlevel;
if(parent = getdad(prsid))
ancprint(parent, indlevel);

if(parent = getmom(prsid))
ancprint(parent, indlevel);

--indlevel;
} /* ancprint */

printgo(ia)
int ia; /* running interactive */
{
long top, bottom;
char cmd;
long save_curp;

save_curp = cur_prs_rowid;
cmd = promptfor("Enter a for an ancestral printout, d for descendents, c for a chain,\n\tor e for everybody:", "cade");
nl();
nl();
top = bottom = 0;

if(cmd != 'e') {
if(cmd != 'a') {
if(ia) top = cur_prs_rowid;
else {
printf("Earliest ancestor:");
top = searchfor(NULL);
nl();
}
}

if(cmd != 'd') {
if(ia && cmd == 'a') bottom = cur_prs_rowid;
else {
printf("Most recent descendent:");
bottom = searchfor(NULL);
nl();
}
}
} /* not everybody */

pf = efopen("printout", "w");
indlevel = 0;
fprintf(pf, "\n\n");

switch(cmd) {
case 'a':
ancprint(bottom);
break;
case 'c':
chainprint(top, bottom);
break;
case 'd':
descprint( top);
break;
case 'e':
everyprint();
break;
} /* end switch */

fclose(pf);
printf("Printout complete, review the file `printout' in your gng directory\n");
if(ia) set_curperson(save_curp);
} /* printgo */

chainprint(top, bottom)
long top, bottom;
{
chainlength = 0;
chaintop = top;
if(!chainup(bottom))
eprint("There is no direct ancestral line connecting these two people");
printf("%d generations\n", chainlength+1);

chainlink[chainlength] = top;
for(; chainlength >= 0; --chainlength) {
set_curperson(chainlink[chainlength]);
topscreen(pf, 0);
}
} /* chainprint */

chainup(prsid)
long prsid;
{
long parent;

if(prsid == chaintop) return 1;
if(chainlength == 200)
eprint("more than 200 generations, cannot search backwards to establish the chain");

if(parent = getdad(prsid)) {
chainlink[chainlength++] = prsid;
if(chainup(parent)) return 1;
--chainlength;
}

if(parent = getmom(prsid)) {
chainlink[chainlength++] = prsid;
if(chainup(parent)) return 1;
--chainlength;
}

return 0;
} /* chainup */

descprint(prsid)
long prsid;
{
int n;

if(indlevel > 50) return; /* no more than 50 generations */
set_curperson(prsid);
topscreen(pf, indlevel);

++indlevel;
spreadout(prsid, 'c');
for(n=0; n<spreadnum; ++n) {
if(spreaddata[n].adopt) continue;
descprint(spreaddata[n].prsid);
/* must rerun spreadout, spreaddata[] has been distroyed */
spreadout(prsid, 'c');
}
--indlevel;
} /* descprint */

everyprint()
{
struct rel_cursor c;
struct person *p;

LOOP_person(&c, p, NULL)
set_curperson(c.rowid);
topscreen(pf, 0);
END_LOOP_person(&c)
} /* everyprint */

