/***********************************************************************
|  Basic calculator with polynomials
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* bcgram.ec: subset of unix bc(1), integer only */

/* This program is not a futile attempt to reinvent UNIX bc(1),
 * or maxima, or any other respectable math package.
 * It is merely a collection of routines and capabilities that I find useful,
 * and I want to have than at hand, rather than logging into some
 * mainframe somewhere.
 * These functions include integer arithmetic, modular arithmetic,
 * crude primality testing and factorization,
 * base conversions (especially hex and decimal),
 * and polynomial manipulations.
 */

global:
eprint_exit set_err_seen;

main:
arguments ".*";

scanner:
string_length 700:
word_length 10:
divert "/* */":
num_fn num_to_poly:
num_huge_fn numh_to_poly:
word_fn word_to_var:

/* last polynomial printed */
token "." DOT DOT 0 dot_to_poly:

/* math operators */
left "-" MINUS ADDOP:
left "+" PLUS ADDOP:
newprec:
left "*" STAR MULOP:
left "%" MOD MULOP:
left "/" SLASH MULOP:
newprec:
left "^" POWER POWER:
left "//" CHOOSE CHOOSE:
newprec:
left "(" LP LP:
token ")" RP RP:
newprec:
prec "!" BANG BANG:
prec "'" PRIME PRIME:

token "@" ROOT ROOT:
token "," COMMA COMMA:
token "=" EQUALS EQUOP:
token "newline" EOL EOL 0 check_nl_quote:
token "\"" QUOTE QUOTE 0 change_quote:

/* simple line commands */
token "quit" LC_QUIT Line_cmd:
token "mod" LC_MOD Line_cmd:
token "ibase" LC_IBASE Line_cmd:
token "obase" LC_OBASE Line_cmd:
token "prime" LC_PRIME Line_cmd:
token "factor" LC_FACTOR Line_cmd:
token "top" LC_TOP Line_cmd:
token "bottom" LC_BOTTOM Line_cmd:
token "sqrt" LC_SQRT Line_cmd:

my_type_ptr POLY;


parser:

rule[input]
|	/* empty file is ok */
|	input line EOL = endline()
:

cast POLY exp NUMBER NUMBER_HUGE:
cast_short WORD:
cast_tok_val Line_cmd:

rule[line]
|	exp = dispexp($1)
| WORD EQUOP exp = assign($1, $3, cast POLY 0)
| WORD ROOT NUMBER EQUOP exp  = assign($1, $5, $3)
|	Line_cmd exp = bc_cmd($1, $2)
|	Line_cmd = bc_cmd($1, cast POLY 0)
|	/* blank line */
|	error
:

cast_tok_val ADDOP MULOP POWER CHOOSE:
cast_long varlist:

rule[exp]
|	ADDOP exp < mathop(cast POLY 0, $1, $2)
|	exp ADDOP exp < mathop($1, $2, $3)
|	exp MULOP exp < mathop($1, $2, $3)
|	exp POWER exp < mathop($1, $2, $3)
|	exp CHOOSE exp < mathop($1, $2, $3)
|	LP exp RP < $2
|	exp BANG < pbang($1)
|	exp PRIME WORD < pderiv($1, $3)
|	exp LP varlist RP < subst($1, $3)
|	WORD < sym_to_poly($1)
|	NUMBER
:

rule[varlist]
|	WORD < addbit(cast_long 0, $1)
|	varlist COMMA WORD < addbit($1, $3)
;


/* end of rules, support routines here. */

${

static char quoteflag = 0; /* in quotes */

set_err_seen() { err_seen = 1; ZC_recflag = 1; }

endline() { allfree(); ZC_recflag = 0; err_seen = 0; }

parse_error(s)
char *s;
{
eprint(s);
return 0;
} /* parse_error */


loopmain_init()
{
scanner_base = 10;
} /* loopmain_init */

loopmain(s)
char *s;
{
if(!pushstream(s))
parse_go();
} /* loopmain */

loopmain_close()
{
if(!pushstream("-"))
parse_go();
} /* loopmain_close */

check_nl_quote(n)
{
if(quoteflag) eprint("missing closing quote");
quoteflag = 0;
return n;
} /* check_nl_quote */

change_quote(n)
{
quoteflag ^= 1;
return quoteflag ? GRAM_LP : GRAM_RP;
} /* change_quote */

dispexp(p)
struct POLY *p;
{
if(err_seen) return;
pdisp(p);
printf("\n");
} /* dispexp */

numh_to_poly()
{
struct POLY *p = newpoly(1);
p->terms[0].coef = string_to_vint(ZC_inp_parseval.s);
ZC_inp_parseval.POLY = p;
return GRAM_NUMBER;
} /* numh_to_poly */

num_to_poly()
{
long l = ZC_inp_parseval.l;
struct POLY *p = &pzero;
if(l) {
p = newpoly(1);
p->terms[0].coef = long_to_vint(l);
}
ZC_inp_parseval.POLY = p;
return GRAM_NUMBER;
} /* num_to_poly */

word_to_var()
{
char *s = ZC_inp_parseval.s;
ZC_inp_parseval.n = 0;

if(!islower(s[0]) || s[1])
eprint("sorry, evolution, only one letter variable names");
else
ZC_inp_parseval.n = *s - 'a';
return GRAM_WORD;
} /* word_to_var */

dot_to_poly()
{
ZC_inp_parseval.POLY = dotval;
return GRAM_NUMBER;
} /* dot_to_poly */

struct POLY *mathop(p1, op, p2)
struct POLY *p1 , *p2;
{
struct POLY *q = &pzero;

switch(op) {
case TOK_PLUS:
if(!p1) p1 = &pzero;
q = radd(p1, p2);
break;
case TOK_MINUS:
if(!p1) p1 = &pzero;
q = rsub(p1, p2);
break;
case TOK_STAR:
q = rmul(p1, p2);
break;
case TOK_SLASH:
q = rdiv(p1, p2);
break;
case TOK_MOD:
q = rrem(p1, p2);
break;
case TOK_POWER:
if(!isconst(p2)) { eprint(er_const); break; }
q = rexp(p1, p2->terms[0].coef);
break;
case TOK_CHOOSE:
if(!isconst(p2)) { eprint(er_const); break; }
q = rchoose(p1, p2->terms[0].coef);
break;
default: eprint("@default[1] %d in bcgram.ec", op);
} /* end switch on math operator */
return q; /* the answer */
} /* mathop */

struct POLY * sym_to_poly(vn)
{
struct POLY *p;

if(!quoteflag && vbr[vn] == 1) {
p = vb[vn];
} else {
p = newpoly(1);
p->terms[0].coef = &ione;
p->terms[0].nvars = 1;
p->terms[0].degree = 1;
p->terms[0].vars[0] = vn;
p->terms[0].exps[0] = 1;
}
return p;
} /* sym_to_poly */

struct POLY *pbang(p)
struct POLY *p;
{
struct POLY *q = &pzero;
struct VINT *t;

if(!isconst(p)) eprint(er_const);
else {
t = p->terms[0].coef;
if(t->size < 0)
eprint("negative factorial");
else if(t->size > 1)
eprint("factorial too large");
else {
t = vbang(t);
q = newpoly(1);
q->terms[0].coef = t;
}
}
return q;
} /* pbang */

long addbit(list, n) long list; { return (1L << n) | list; }
$}
